"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgNotifyClient = void 0;
const common_1 = require("@nestjs/common");
const microservices_1 = require("@nestjs/microservices");
const pg_listen_1 = require("pg-listen");
const pg_notify_response_1 = require("./pg-notify.response");
const pg_notify_util_1 = require("./pg-notify.util");
class PgNotifyClient extends microservices_1.ClientProxy {
    constructor(options) {
        super();
        this.subscriptionsCount = new Map();
        this.loggerService = options.logger || new common_1.Logger();
        this.publisher = this.createClient(options);
        this.firstConnected = false;
        this.connected = false;
    }
    async onApplicationBootstrap() {
        await this.connectOnBootstrap();
    }
    async onApplicationShutdown() {
        await this.close();
    }
    async connectOnBootstrap() {
        try {
            this.firstConnected = true;
            await this.publisher.connect();
            this.bindMessageHandlers();
        }
        catch (error) {
            this.firstConnected = false;
            this.publisher.events.emit('error', error);
        }
    }
    async connect() {
        if (!this.connected) {
            throw new Error('Client is not connected');
        }
    }
    async close() {
        try {
            await this.publisher.unlistenAll();
            await this.publisher.close();
        }
        catch (error) {
            this.publisher.events.emit('error', error);
        }
    }
    async dispatchEvent(packet) {
        const pattern = this.normalizePattern(packet.pattern);
        await this.publisher.notify(pattern, packet.data);
    }
    publish(partialPacket, callback) {
        try {
            const packet = this.assignPacketId(partialPacket);
            const pattern = this.normalizePattern(partialPacket.pattern);
            const responseChannel = pg_notify_util_1.getReplyPattern(pattern);
            let subscriptionsCount = this.subscriptionsCount.get(responseChannel) || 0;
            const publishPacket = async () => {
                subscriptionsCount = this.subscriptionsCount.get(responseChannel) || 0;
                this.subscriptionsCount.set(responseChannel, subscriptionsCount + 1);
                this.routingMap.set(packet.id, callback);
                const payload = { id: packet.id, data: packet.data };
                await this.publisher.notify(pattern, JSON.stringify(payload));
            };
            if (subscriptionsCount <= 0) {
                this.publisher.listenTo(responseChannel);
            }
            publishPacket()
                .then()
                .catch(err => callback({ err }));
            return () => {
                this.unsubscribeFromChannel(responseChannel);
                this.routingMap.delete(packet.id);
            };
        }
        catch (err) {
            return callback({ err });
        }
    }
    createClient(options) {
        var _a, _b, _c;
        const client = pg_listen_1.default(options.connection, {
            retryInterval: (_a = options.strategy) === null || _a === void 0 ? void 0 : _a.retryInterval,
            retryLimit: (_b = options.strategy) === null || _b === void 0 ? void 0 : _b.retryLimit,
            retryTimeout: (_c = options.strategy) === null || _c === void 0 ? void 0 : _c.retryTimeout,
        });
        this.bindEventHandlers(client);
        return client;
    }
    bindMessageHandlers() {
        this.publisher.events.on('notification', this.createResponseCallback());
    }
    bindEventHandlers(publisher) {
        publisher.events.on('connected', () => {
            if (!this.firstConnected) {
                this.bindMessageHandlers();
            }
            this.connected = true;
            this.firstConnected = true;
            this.loggerService.log('Connection established', PgNotifyClient.name);
        });
        publisher.events.on('error', error => {
            const defaultMessage = 'Internal error';
            const message = error.message || defaultMessage;
            this.loggerService.error(message, error.stack, PgNotifyClient.name);
        });
        publisher.events.on('reconnect', attempt => {
            this.connected = false;
            this.loggerService.error(`Connection refused. Retry attempt ${attempt}...`, undefined, PgNotifyClient.name);
        });
    }
    createResponseCallback() {
        return (notification) => {
            const packet = JSON.parse(notification.payload);
            const { err, response, isDisposed, id } = packet;
            const parsedResponse = this.getResponse(response);
            const callback = this.routingMap.get(id);
            if (!callback)
                return;
            if (isDisposed || err) {
                return callback({ err, response: parsedResponse, isDisposed: true });
            }
            return callback({ err, response: parsedResponse });
        };
    }
    unsubscribeFromChannel(channel) {
        const subscriptionCount = this.subscriptionsCount.get(channel);
        if (subscriptionCount) {
            this.subscriptionsCount.set(channel, subscriptionCount - 1);
            if (subscriptionCount - 1 <= 0) {
                this.publisher.unlisten(channel);
            }
        }
    }
    getResponse(response) {
        if (pg_notify_util_1.isObject(response) && response.status) {
            return new pg_notify_response_1.PgNotifyResponse(response.status, response.data, response.error);
        }
        return response;
    }
}
exports.PgNotifyClient = PgNotifyClient;
