"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgNotifyServer = void 0;
const common_1 = require("@nestjs/common");
const microservices_1 = require("@nestjs/microservices");
const constants_1 = require("@nestjs/microservices/constants");
const pg_listen_1 = require("pg-listen");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const pg_notify_constant_1 = require("./pg-notify.constant");
const pg_notify_context_1 = require("./pg-notify.context");
const pg_notify_response_1 = require("./pg-notify.response");
const pg_notify_util_1 = require("./pg-notify.util");
class PgNotifyServer extends microservices_1.Server {
    constructor(options) {
        super();
        this.loggerService = options.logger || new common_1.Logger();
        this.subscriber = this.createClient(options);
        this.firstConnected = false;
    }
    async listen(callback) {
        try {
            this.firstConnected = true;
            await this.subscriber.connect();
            await this.listenChannels();
            this.bindMessageHandlers();
        }
        catch (err) {
            this.firstConnected = false;
            this.subscriber.events.emit('error', err);
        }
        finally {
            callback();
        }
    }
    async close() {
        try {
            await this.subscriber.unlistenAll();
            await this.subscriber.close();
        }
        catch (err) {
            this.subscriber.events.emit('error', err);
        }
    }
    addHandler(pattern, callback) {
        if (pg_notify_util_1.isObject(pattern) && pattern.transport === pg_notify_constant_1.PG_NOTIFY_TRANSPORT) {
            super.addHandler(pattern.pattern, callback, pattern.isEvent);
        }
    }
    createClient(options) {
        var _a, _b, _c;
        const subscriber = pg_listen_1.default(options.connection, {
            retryInterval: (_a = options.strategy) === null || _a === void 0 ? void 0 : _a.retryInterval,
            retryLimit: (_b = options.strategy) === null || _b === void 0 ? void 0 : _b.retryLimit,
            retryTimeout: (_c = options.strategy) === null || _c === void 0 ? void 0 : _c.retryTimeout,
        });
        this.bindEventHandlers(subscriber);
        return subscriber;
    }
    async listenChannels() {
        const channels = Array.from(this.messageHandlers.keys());
        await Promise.all(channels.map(channel => this.subscriber.listenTo(channel)));
    }
    bindMessageHandlers() {
        this.subscriber.events.on('notification', async (notification) => {
            const payload = this.parsePayload(notification.payload);
            const ctx = new pg_notify_context_1.PgNotifyContext(notification.processId, notification.channel, payload.data, payload.id);
            if (payload.id) {
                return this.handleAsRequest(notification.channel, payload.id, payload.data, ctx);
            }
            return this.handleAsEvent(notification.channel, payload.data, ctx);
        });
    }
    bindEventHandlers(subscriber) {
        subscriber.events.on('connected', async () => {
            if (!this.firstConnected) {
                try {
                    await this.listenChannels();
                    this.bindMessageHandlers();
                    this.firstConnected = true;
                }
                catch (err) {
                    this.subscriber.events.emit('error', err);
                }
            }
            this.loggerService.log('Connection established', PgNotifyServer.name);
        });
        subscriber.events.on('error', err => {
            const defaultMessage = 'Internal error';
            const message = pg_notify_util_1.parseErrorMessage(err) || defaultMessage;
            this.loggerService.error(message, err.stack, PgNotifyServer.name);
        });
        subscriber.events.on('reconnect', attempt => {
            this.loggerService.error(`Connection refused. Retry attempt ${attempt}...`, undefined, PgNotifyServer.name);
        });
    }
    async handleAsEvent(channel, data, ctx) {
        const handler = this.getHandlerByPattern(channel);
        if (!handler) {
            return this.loggerService.error(`${constants_1.NO_EVENT_HANDLER} Event pattern: ${channel}.`, undefined, PgNotifyServer.name);
        }
        try {
            const resolvedHandler = await handler(data, ctx);
            const stream = this
                .transformToObservable(resolvedHandler)
                .pipe(operators_1.publish());
            stream.connect();
        }
        catch (err) {
            return this.loggerService.error(pg_notify_util_1.parseErrorMessage(err), undefined, PgNotifyServer.name);
        }
    }
    async handleAsRequest(channel, id, data, ctx) {
        const handler = this.getHandlerByPattern(channel);
        const publisher = this.getPublisher(channel, id);
        if (!handler) {
            const response$ = rxjs_1.of(pg_notify_response_1.PgNotifyResponse.error(constants_1.NO_MESSAGE_HANDLER, 404));
            return this.send(response$, publisher);
        }
        try {
            const resolvedHandler = await handler(data, ctx);
            const response$ = this.transformToObservable(resolvedHandler).pipe(operators_1.map(res => this.getResponse(res)), operators_1.catchError(err => rxjs_1.of(pg_notify_response_1.PgNotifyResponse.error(pg_notify_util_1.parseErrorMessage(err)))));
            return this.send(response$, publisher);
        }
        catch (err) {
            const response$ = rxjs_1.of(pg_notify_response_1.PgNotifyResponse.error(pg_notify_util_1.parseErrorMessage(err)));
            return this.send(response$, publisher);
        }
    }
    getPublisher(pattern, id) {
        return async (response) => {
            Object.assign(response, { id });
            await this.subscriber.notify(pg_notify_util_1.getReplyPattern(pattern), JSON.stringify(response));
        };
    }
    parsePayload(payload) {
        const parsedPayload = { id: undefined, data: undefined };
        try {
            const payloadAsObject = JSON.parse(payload);
            if (payloadAsObject.data) {
                parsedPayload.id = payloadAsObject.id;
                parsedPayload.data = payloadAsObject.data;
            }
            else {
                parsedPayload.data = payloadAsObject;
            }
        }
        catch (err) {
            parsedPayload.data = payload;
        }
        return parsedPayload;
    }
    getResponse(data) {
        if (data instanceof pg_notify_response_1.PgNotifyResponse) {
            return data;
        }
        return pg_notify_response_1.PgNotifyResponse.success(data);
    }
}
exports.PgNotifyServer = PgNotifyServer;
